@@ -29,6 +29,7 @@ class Commands(discord.Cog, name="Commands"):
         index_cog,
         translations_cog=None,
         search_cog=None,
+        transcribe_cog=None,
     ):
         super().__init__()
         self.bot = bot
@@ -43,6 +44,7 @@ class Commands(discord.Cog, name="Commands"):
         self.index_cog = index_cog
         self.translations_cog = translations_cog
         self.search_cog = search_cog
+        self.transcribe_cog = transcribe_cog
 
     # Create slash command groups
     dalle = discord.SlashCommandGroup(
@@ -75,6 +77,18 @@ class Commands(discord.Cog, name="Commands"):
         guild_ids=ALLOWED_GUILDS,
         checks=[Check.check_index_roles()],
     )
+    transcribe = discord.SlashCommandGroup(
+        name="transcribe",
+        description="Transcription services using OpenAI Whisper",
+        guild_ids=ALLOWED_GUILDS,
+        checks=[Check.check_index_roles()],  # TODO new role checker for transcribe
+    )
+    internet = discord.SlashCommandGroup(
+        name="internet",
+        description="Transcription services using OpenAI Whisper2",
+        guild_ids=ALLOWED_GUILDS,
+        checks=[Check.check_index_roles()],  # TODO new role checker for transcribe
+    )
 
     #
     # System commands
@@ -302,12 +316,68 @@ class Commands(discord.Cog, name="Commands"):
 
     @add_to_group("gpt")
     @discord.slash_command(
-        name="ask",
-        description="Ask GPT3 something!",
+        name="instruction",
+        description="Set your own system instruction",
         guild_ids=ALLOWED_GUILDS,
     )
     @discord.option(
-        name="prompt", description="The prompt to send to GPT3", required=True
+        name="mode",
+        description="Set/Get/Clear prompt",
+        choices=["set", "get", "clear"],
+        required=True,
+    )
+    @discord.option(
+        name="type",
+        description="Enable for channel or for user",
+        choices=["user", "channel"],
+        required=True,
+    )
+    @discord.option(
+        name="instruction", description="The instruction to set", required=False
+    )
+    @discord.option(
+        name="instruction_file",
+        description="The instruction to set from a txt file",
+        input_type=discord.SlashCommandOptionType.attachment,
+        required=False,
+    )
+    @discord.option(
+        name="private", description="Will only be visible to you", required=False
+    )
+    @discord.guild_only()
+    async def instruction(
+        self,
+        ctx: discord.ApplicationContext,
+        mode: str,
+        type: str,
+        instruction: str,
+        instruction_file: discord.Attachment,
+        private: bool,
+    ):
+        await self.converser_cog.instruction_command(
+            ctx, mode, type, instruction, instruction_file, private
+        )
+
+    @add_to_group("gpt")
+    @discord.slash_command(
+        name="ask",
+        description="Ask the bot something!",
+        guild_ids=ALLOWED_GUILDS,
+    )
+    @discord.option(
+        name="prompt", description="The prompt to send to the model", required=False
+    )
+    @discord.option(
+        name="prompt_file",
+        description="The prompt file to send to the model. Is added before the prompt, both can be combined",
+        required=False,
+        input_type=discord.SlashCommandOptionType.attachment,
+    )
+    @discord.option(
+        name="model",
+        description="The model to use for the request",
+        required=False,
+        autocomplete=Settings_autocompleter.get_models,
     )
     @discord.option(
         name="private", description="Will only be visible to you", required=False
@@ -345,6 +415,8 @@ class Commands(discord.Cog, name="Commands"):
         self,
         ctx: discord.ApplicationContext,
         prompt: str,
+        prompt_file: discord.Attachment,
+        model: str,
         private: bool,
         temperature: float,
         top_p: float,
@@ -359,17 +431,19 @@ class Commands(discord.Cog, name="Commands"):
             top_p,
             frequency_penalty,
             presence_penalty,
+            prompt_file=prompt_file,
+            model=model,
         )
 
     @add_to_group("gpt")
     @discord.slash_command(
         name="edit",
-        description="Ask GPT3 to edit some text!",
+        description="Ask the bot to edit some text!",
         guild_ids=ALLOWED_GUILDS,
     )
     @discord.option(
         name="instruction",
-        description="How you want GPT3 to edit the text",
+        description="How you want the bot to edit the text",
         required=True,
     )
     @discord.option(
@@ -397,9 +471,6 @@ class Commands(discord.Cog, name="Commands"):
         min_value=0,
         max_value=1,
     )
-    @discord.option(
-        name="codex", description="Enable codex version", required=False, default=False
-    )
     @discord.guild_only()
     async def edit(
         self,
@@ -409,16 +480,15 @@ class Commands(discord.Cog, name="Commands"):
         private: bool,
         temperature: float,
         top_p: float,
-        codex: bool,
     ):
         await self.converser_cog.edit_command(
-            ctx, instruction, text, private, temperature, top_p, codex
+            ctx, instruction, text, private, temperature, top_p
         )
 
     @add_to_group("gpt")
     @discord.slash_command(
         name="converse",
-        description="Have a conversation with GPT3",
+        description="Have a conversation with GPT",
         guild_ids=ALLOWED_GUILDS,
     )
     @discord.option(
@@ -449,7 +519,7 @@ class Commands(discord.Cog, name="Commands"):
         description="Which model to use with the bot",
         required=False,
         default=False,
-        autocomplete=Settings_autocompleter.get_models,
+        autocomplete=Settings_autocompleter.get_converse_models,
     )
     @discord.option(
         name="temperature",
@@ -483,6 +553,12 @@ class Commands(discord.Cog, name="Commands"):
         min_value=-2,
         max_value=2,
     )
+    @discord.option(
+        name="use_threads",
+        description="Set this to false to start a channel conversation",
+        required=False,
+        default=True,
+    )
     @discord.guild_only()
     async def converse(
         self,
@@ -496,6 +572,7 @@ class Commands(discord.Cog, name="Commands"):
         top_p: float,
         frequency_penalty: float,
         presence_penalty: float,
+        use_threads: bool,
     ):
         await self.converser_cog.converse_command(
             ctx,
@@ -508,12 +585,13 @@ class Commands(discord.Cog, name="Commands"):
             top_p,
             frequency_penalty,
             presence_penalty,
+            use_threads=use_threads,
         )
 
     @add_to_group("gpt")
     @discord.slash_command(
         name="end",
-        description="End a conversation with GPT3",
+        description="End a conversation with GPT",
         guild_ids=ALLOWED_GUILDS,
     )
     @discord.guild_only()
@@ -644,6 +722,42 @@ class Commands(discord.Cog, name="Commands"):
             ctx, user_index, server_index, search_index
         )
 
+    @add_to_group("index")
+    @discord.slash_command(
+        name="talk",
+        description="Select one of your saved indexes to talk to",
+        guild_ids=ALLOWED_GUILDS,
+    )
+    @discord.guild_only()
+    @discord.option(
+        name="user_index",
+        description="Which user file to load the index from",
+        required=False,
+        autocomplete=File_autocompleter.get_user_indexes,
+    )
+    @discord.option(
+        name="search_index",
+        description="Which search index file to load the index from",
+        required=False,
+        autocomplete=File_autocompleter.get_user_search_indexes,
+    )
+    @discord.option(
+        name="model",
+        description="The model to use for the conversation",
+        required=False,
+        default="gpt-3.5-turbo",
+        autocomplete=Settings_autocompleter.get_index_and_search_models,
+    )
+    async def talk(
+        self,
+        ctx: discord.ApplicationContext,
+        user_index: str,
+        search_index: str,
+        model: str,
+    ):
+        await ctx.defer()
+        await self.index_cog.index_chat_command(ctx, user_index, search_index, model)
+
     @add_to_group("index")
     @discord.slash_command(
         name="add", description="Add an index to query from", guild_ids=ALLOWED_GUILDS
@@ -666,6 +780,32 @@ class Commands(discord.Cog, name="Commands"):
     ):
         await self.index_cog.set_index_command(ctx, file, link)
 
+    @add_to_group("index")
+    @discord.slash_command(
+        name="recurse-link",
+        description="Recursively index a link",
+        guild_ids=ALLOWED_GUILDS,
+    )
+    @discord.guild_only()
+    @discord.option(
+        name="link",
+        description="A link to create the index from",
+        required=True,
+        input_type=discord.SlashCommandOptionType.string,
+    )
+    @discord.option(
+        name="depth",
+        description="How deep to recurse",
+        required=False,
+        input_type=discord.SlashCommandOptionType.integer,
+        min_value=1,
+        max_value=5,
+    )
+    async def set_recurse_link(
+        self, ctx: discord.ApplicationContext, link: str, depth: int
+    ):
+        await self.index_cog.set_index_link_recurse_command(ctx, link, depth)
+
     @add_to_group("index")
     @discord.slash_command(
         name="reset",
@@ -758,8 +898,8 @@ class Commands(discord.Cog, name="Commands"):
         description="Response mode, doesn't work on deep composed indexes",
         guild_ids=ALLOWED_GUILDS,
         required=False,
-        default="default",
-        choices=["default", "compact", "tree_summarize"],
+        default="refine",
+        choices=["refine", "compact", "tree_summarize"],
     )
     @discord.option(
         name="child_branch_factor",
@@ -770,6 +910,20 @@ class Commands(discord.Cog, name="Commands"):
         max_value=3,
         input_type=discord.SlashCommandOptionType.integer,
     )
+    @discord.option(
+        name="model",
+        description="The model to use for the request (querying, not composition)",
+        required=False,
+        default="gpt-3.5-turbo",
+        autocomplete=Settings_autocompleter.get_index_and_search_models,
+    )
+    @discord.option(
+        name="multistep",
+        description="Do a more intensive, multi-step query,",
+        required=False,
+        default=False,
+        input_type=discord.SlashCommandOptionType.boolean,
+    )
     async def query(
         self,
         ctx: discord.ApplicationContext,
@@ -777,10 +931,18 @@ class Commands(discord.Cog, name="Commands"):
         nodes: int,
         response_mode: str,
         child_branch_factor: int,
+        model: str,
+        multistep: bool,
     ):
         await ctx.defer()
         await self.index_cog.query_command(
-            ctx, query, nodes, response_mode, child_branch_factor
+            ctx,
+            query,
+            nodes,
+            response_mode,
+            child_branch_factor,
+            model,
+            multistep,
         )
 
     #
@@ -951,10 +1113,44 @@ class Commands(discord.Cog, name="Commands"):
     async def summarize_action(self, ctx, message: discord.Message):
         await self.converser_cog.summarize_action(ctx, message)
 
+    @add_to_group("internet")
+    @discord.slash_command(
+        name="chat",
+        description="Chat with GPT connected to the internet!",
+        guild_ids=ALLOWED_GUILDS,
+        checks=[Check.check_search_roles()],
+    )
+    @discord.option(
+        name="search_scope",
+        description="How many top links to use for context",
+        required=False,
+        input_type=discord.SlashCommandOptionType.integer,
+        max_value=6,
+        min_value=1,
+        default=2,
+    )
+    @discord.option(
+        name="model",
+        description="The model to use for the request (querying, not composition)",
+        required=False,
+        default="gpt-3.5-turbo",
+        autocomplete=Settings_autocompleter.get_index_and_search_models,
+    )
+    async def chat(
+        self,
+        ctx: discord.ApplicationContext,
+        model: str,
+        search_scope: int = 2,
+    ):
+        await self.search_cog.search_chat_command(
+            ctx, search_scope=search_scope, model=model
+        )
+
     # Search slash commands
+    @add_to_group("internet")
     @discord.slash_command(
         name="search",
-        description="Search google alongside GPT3 for something",
+        description="Search google alongside GPT for something",
         guild_ids=ALLOWED_GUILDS,
         checks=[Check.check_search_roles()],
     )
@@ -986,8 +1182,22 @@ class Commands(discord.Cog, name="Commands"):
         description="Response mode, doesn't work on deep searches",
         guild_ids=ALLOWED_GUILDS,
         required=False,
-        default="default",
-        choices=["default", "compact", "tree_summarize"],
+        default="refine",
+        choices=["refine", "compact", "tree_summarize"],
+    )
+    @discord.option(
+        name="model",
+        description="The model to use for the request (querying, not composition)",
+        required=False,
+        default="gpt-3.5-turbo",
+        autocomplete=Settings_autocompleter.get_index_and_search_models,
+    )
+    @discord.option(
+        name="multistep",
+        description="Do a more intensive, multi-step query,",
+        required=False,
+        default=False,
+        input_type=discord.SlashCommandOptionType.boolean,
     )
     @discord.guild_only()
     async def search(
@@ -998,7 +1208,72 @@ class Commands(discord.Cog, name="Commands"):
         nodes: int,
         deep: bool,
         response_mode: str,
+        model: str,
+        multistep: bool,
     ):
         await self.search_cog.search_command(
-            ctx, query, scope, nodes, deep, response_mode
+            ctx,
+            query,
+            scope,
+            nodes,
+            deep,
+            response_mode,
+            model,
+            multistep,
         )
+
+    # Transcribe commands
+    @add_to_group("transcribe")
+    @discord.slash_command(
+        name="file",
+        description="Transcribe an audio or video file",
+        guild_ids=ALLOWED_GUILDS,
+    )
+    @discord.guild_only()
+    @discord.option(
+        name="file",
+        description="A file to transcribe",
+        required=True,
+        input_type=discord.SlashCommandOptionType.attachment,
+    )
+    @discord.option(
+        name="temperature",
+        description="The higher the value, the riskier the model will be",
+        required=False,
+        input_type=discord.SlashCommandOptionType.number,
+        max_value=1,
+        min_value=0,
+    )
+    async def transcribe_file(
+        self,
+        ctx: discord.ApplicationContext,
+        file: discord.Attachment,
+        temperature: float,
+    ):
+        await self.transcribe_cog.transcribe_file_command(ctx, file, temperature)
+
+    @add_to_group("transcribe")
+    @discord.slash_command(
+        name="link",
+        description="Transcribe a file link or youtube link",
+        guild_ids=ALLOWED_GUILDS,
+    )
+    @discord.guild_only()
+    @discord.option(
+        name="link",
+        description="A link to transcribe",
+        required=True,
+        input_type=discord.SlashCommandOptionType.string,
+    )
+    @discord.option(
+        name="temperature",
+        description="The higher the value, the riskier the model will be",
+        required=False,
+        input_type=discord.SlashCommandOptionType.number,
+        max_value=1,
+        min_value=0,
+    )
+    async def transcribe_link(
+        self, ctx: discord.ApplicationContext, link: str, temperature: float
+    ):
+        await self.transcribe_cog.transcribe_link_command(ctx, link, temperature)
