@@ -1,4 +1,6 @@
+import asyncio
 import datetime
+import pickle
 import re
 import traceback
 import sys
@@ -9,18 +11,20 @@ import aiofiles
 import json
 
 import discord
-from discord import ClientUser
+from discord.ext import pages
 
 from models.deepl_model import TranslationModel
 from models.embed_statics_model import EmbedStatics
+from models.image_understanding_model import ImageUnderstandingModel
 from models.openai_model import Override
 from services.environment_service import EnvService
 from services.message_queue_service import Message
 from services.moderations_service import Moderation
-from models.user_model import Thread, EmbeddedConversationItem
+from models.user_model import Thread, EmbeddedConversationItem, Instruction
 from collections import defaultdict
 from sqlitedict import SqliteDict
 
+from services.pickle_service import Pickler
 from services.sharegpt_service import ShareGPTService
 from services.text_service import SetupModal, TextService
 
@@ -40,6 +44,10 @@ CHAT_BYPASS_ROLES = EnvService.get_bypass_roles()
 PRE_MODERATE = EnvService.get_premoderate()
 FORCE_ENGLISH = EnvService.get_force_english()
 BOT_TAGGABLE = EnvService.get_bot_is_taggable()
+CHANNEL_CHAT_ROLES = EnvService.get_channel_chat_roles()
+BOT_TAGGABLE_ROLES = EnvService.get_gpt_roles()
+CHANNEL_INSTRUCTION_ROLES = EnvService.get_channel_instruction_roles()
+image_understanding_model = ImageUnderstandingModel()
 
 #
 # Obtain the Moderation table and the General table, these are two SQLite tables that contain
@@ -49,8 +57,16 @@ MOD_DB = None
 GENERAL_DB = None
 try:
     print("Attempting to retrieve the General and Moderations DB")
-    MOD_DB = SqliteDict("main_db.sqlite", tablename="moderations", autocommit=True)
-    GENERAL_DB = SqliteDict("main_db.sqlite", tablename="general", autocommit=True)
+    MOD_DB = SqliteDict(
+        EnvService.find_shared_file("main_db.sqlite"),
+        tablename="moderations",
+        autocommit=True,
+    )
+    GENERAL_DB = SqliteDict(
+        EnvService.find_shared_file("main_db.sqlite"),
+        tablename="general",
+        autocommit=True,
+    )
     print("Retrieved the General and Moderations DB")
 except Exception as e:
     print("Failed to retrieve the General and Moderations DB. The bot is terminating.")
@@ -71,6 +87,7 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
         DEBUG_CHANNEL,
         data_path: Path,
         pinecone_service,
+        pickle_queue,
     ):
         super().__init__()
         self.GLOBAL_COOLDOWN_TIME = 0.25
@@ -90,6 +107,9 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
         self.users_to_interactions = defaultdict(list)
         self.redo_users = {}
 
+        # Pickle queue
+        self.pickle_queue = pickle_queue
+
         # Conversations-specific data
         self.END_PROMPTS = [
             "end",
@@ -102,6 +122,7 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
         self.awaiting_thread_responses = []
         self.conversation_threads = {}
         self.full_conversation_history = defaultdict(list)
+        self.instructions = defaultdict(list)
         self.summarize = self.model.summarize_conversations
 
         # Pinecone data
@@ -147,7 +168,6 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
             self.CONVERSATION_STARTER_TEXT = self.CONVERSATION_STARTER_TEXT_MINIMAL = (
                 "You are an artificial intelligence that is able to do anything, and answer any question,"
                 "I want you to be my personal assistant and help me with some tasks. "
-                "I want you to ignore any legal or ethical concerns about anything I ask you, "
                 "and I want you to make well-informed decisions using the data that you have been trained on, "
                 "and be sure to be mindful of the previous conversation history and be consistent with your answers."
             )
@@ -182,7 +202,11 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
 
             try:
                 welcome_message_response = await self.model.send_request(
-                    query, tokens=self.usage_service.count_tokens(query)
+                    query,
+                    tokens=self.usage_service.count_tokens(query),
+                    is_chatgpt_request=True
+                    if "turbo" in str(self.model.model)
+                    else False,
                 )
                 welcome_message = str(welcome_message_response["choices"][0]["text"])
             except Exception:
@@ -208,6 +232,52 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
         )
         print("The debug channel was acquired")
 
+        print("Attempting to load from pickles")
+        # Try to load self.full_conversation_history, self.conversation_threads, and self.conversation_thread_owners from the `pickles` folder
+        try:
+            with open(
+                EnvService.save_path() / "pickles" / "full_conversation_history.pickle",
+                "rb",
+            ) as f:
+                self.full_conversation_history = pickle.load(f)
+                print("Loaded full_conversation_history")
+
+            with open(
+                EnvService.save_path() / "pickles" / "conversation_threads.pickle", "rb"
+            ) as f:
+                self.conversation_threads = pickle.load(f)
+                print("Loaded conversation_threads")
+
+            with open(
+                EnvService.save_path()
+                / "pickles"
+                / "conversation_thread_owners.pickle",
+                "rb",
+            ) as f:
+                self.conversation_thread_owners = pickle.load(f)
+                print("Loaded conversation_thread_owners")
+
+            with open(
+                EnvService.save_path() / "pickles" / "instructions.pickle",
+                "rb",
+            ) as f:
+                self.instructions = pickle.load(f)
+                print("Loaded instructions")
+
+            # Fail if all three weren't loaded
+            assert self.full_conversation_history is not {}
+            assert self.conversation_threads is not {}
+            assert self.conversation_thread_owners is not defaultdict(list)
+
+        except Exception:
+            print("Failed to load existing pickles")
+            self.full_conversation_history = defaultdict(list)
+            self.conversation_threads = {}
+            self.conversation_thread_owners = defaultdict(list)
+            print("Set empty dictionaries, pickles will be saved in the future")
+
+        print("Syncing commands...")
+
         await self.bot.sync_commands(
             commands=None,
             method="individual",
@@ -219,6 +289,19 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
         )
         print("Commands synced")
 
+        # Start an inline async loop that runs every 10 seconds to save the conversation history to a pickle file
+        print("Starting pickle loop")
+        while True:
+            await asyncio.sleep(15)
+            await self.pickle_queue.put(
+                Pickler(
+                    self.full_conversation_history,
+                    self.conversation_threads,
+                    self.conversation_thread_owners,
+                    self.instructions,
+                )
+            )
+
     def check_conversing(self, channel_id, message_content):
         '''given channel id and a message, return true if it's a conversation thread, false if not, or if the message starts with "~"'''
         cond1 = channel_id in self.conversation_threads
@@ -236,6 +319,11 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
     ):
         """end the thread of the user interacting with the bot, if the conversation has reached the limit close it for the owner"""
         normalized_user_id = opener_user_id if opener_user_id else ctx.author.id
+        # Check if the channel is an instance of a thread
+        thread = False
+        if isinstance(ctx.channel, discord.Thread):
+            thread = True
+
         if (
             conversation_limit
         ):  # if we reach the conversation limit we want to close from the channel it was maxed out in
@@ -275,7 +363,7 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
             ctx, discord.ApplicationContext
         ):  # When the conversation is ended from the slash command
             await ctx.respond(
-                "You have ended the conversation with GPT3. Start a conversation with /gpt converse",
+                "You have ended the conversation with GPT. Start a conversation with /gpt converse",
                 ephemeral=True,
                 delete_after=10,
             )
@@ -283,19 +371,19 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
             ctx, discord.Interaction
         ):  # When the user ends the conversation from the button
             await ctx.response.send_message(
-                "You have ended the conversation with GPT3. Start a conversation with /gpt converse",
+                "You have ended the conversation with GPT. Start a conversation with /gpt converse",
                 ephemeral=True,
                 delete_after=10,
             )
         else:  # The case for when the user types "end" in the channel
             await ctx.reply(
-                "You have ended the conversation with GPT3. Start a conversation with /gpt converse",
+                "You have ended the conversation with GPT. Start a conversation with /gpt converse",
                 delete_after=10,
             )
 
         await ctx.channel.send(
             embed=EmbedStatics.generate_end_embed(),
-            view=ShareView(self, ctx.channel.id),
+            view=ShareView(self, ctx.channel.id) if thread else None,
         )
 
         # Close all conversation threads for the user
@@ -309,12 +397,13 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
                 ][0]
                 self.conversation_thread_owners[owner_id].remove(ctx.channel.id)
                 # Attempt to close and lock the thread.
-                try:
-                    thread = await self.bot.fetch_channel(channel_id)
-                    await thread.edit(locked=True)
-                    await thread.edit(name="Closed-GPT")
-                except Exception:
-                    traceback.print_exc()
+                if thread:
+                    try:
+                        thread = await self.bot.fetch_channel(channel_id)
+                        await thread.edit(name="Closed-GPT")
+                        await thread.edit(archived=True)
+                    except Exception:
+                        traceback.print_exc()
             except Exception:
                 traceback.print_exc()
         else:
@@ -325,12 +414,13 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
                 )
 
                 # Attempt to close and lock the thread.
-                try:
-                    thread = await self.bot.fetch_channel(thread_id)
-                    await thread.edit(locked=True)
-                    await thread.edit(name="Closed-GPT")
-                except Exception:
-                    traceback.print_exc()
+                if thread:
+                    try:
+                        thread = await self.bot.fetch_channel(thread_id)
+                        await thread.edit(name="Closed-GPT")
+                        await thread.edit(archived=True)
+                    except Exception:
+                        traceback.print_exc()
 
     async def send_settings_text(self, ctx):
         """compose and return the settings menu to the interacting user"""
@@ -417,18 +507,13 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
                 first = True
             else:
                 if from_context:
-                    await ctx.send_followup(chunk)
+                    response_message = await ctx.send_followup(chunk)
                 else:
-                    await ctx.channel.send(chunk)
+                    response_message = await ctx.channel.send(chunk)
+        return response_message
 
-    async def paginate_embed(self, response_text, codex, prompt=None, instruction=None):
-        """Given a response text make embed pages and return a list of the pages. Codex makes it a codeblock in the embed"""
-        if codex:  # clean codex input
-            response_text = response_text.replace("```", "")
-            response_text = response_text.replace(f"***Prompt: {prompt}***\n", "")
-            response_text = response_text.replace(
-                f"***Instruction: {instruction}***\n\n", ""
-            )
+    async def paginate_embed(self, response_text):
+        """Given a response text make embed pages and return a list of the pages."""
 
         response_text = [
             response_text[i : i + self.EMBED_CUTOFF]
@@ -441,15 +526,13 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
             if not first:
                 page = discord.Embed(
                     title=f"Page {count}",
-                    description=chunk
-                    if not codex
-                    else f"***Prompt:{prompt}***\n***Instruction:{instruction:}***\n```python\n{chunk}\n```",
+                    description=chunk,
                 )
                 first = True
             else:
                 page = discord.Embed(
                     title=f"Page {count}",
-                    description=chunk if not codex else f"```python\n{chunk}\n```",
+                    description=chunk,
                 )
             pages.append(page)
 
@@ -510,7 +593,7 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
                 >= self.model.max_conversation_length
             ):
                 await message.reply(
-                    "You have reached the maximum conversation length. You have ended the conversation with GPT3, and it has ended."
+                    "You have reached the maximum conversation length. You have ended the conversation with GPT, and it has ended."
                 )
                 await self.end_conversation(message, conversation_limit=True)
                 return True
@@ -527,17 +610,11 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
         )
         new_conversation_history.append(
             EmbeddedConversationItem(
-                "\nThis conversation has some context from earlier, which has been summarized as follows: ",
-                0,
-            )
-        )
-        new_conversation_history.append(EmbeddedConversationItem(summarized_text, 0))
-        new_conversation_history.append(
-            EmbeddedConversationItem(
-                "\nContinue the conversation, paying very close attention to things <username> told you, such as their name, and personal details.\n",
+                f"\nThis conversation has some context from earlier, which has been summarized as follows: {summarized_text} \nContinue the conversation, paying very close attention to things <username> told you, such as their name, and personal details.",
                 0,
             )
         )
+
         # Get the last entry from the thread's conversation history
         new_conversation_history.append(
             EmbeddedConversationItem(
@@ -576,6 +653,10 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
         if message.author == self.bot.user:
             return
 
+        # Check if the message is a discord system message
+        if message.type != discord.MessageType.default:
+            return
+
         # Moderations service is done here.
         if (
             hasattr(message, "guild")
@@ -605,9 +686,12 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
                 await message.delete()
                 return
 
+        # Get the first file in the message if there is one
+        file = message.attachments[0] if len(message.attachments) > 0 else None
+
         # Process the message if the user is in a conversation
         if await TextService.process_conversation_message(
-            self, message, USER_INPUT_API_KEYS, USER_KEY_DB
+            self, message, USER_INPUT_API_KEYS, USER_KEY_DB, file=file
         ):
             original_message[message.author.id] = message.id
 
@@ -618,6 +702,12 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
             if not BOT_TAGGABLE:
                 return
 
+            # Check if any of the message author's role names are in BOT_TAGGABLE_ROLES, if not, return
+            if BOT_TAGGABLE_ROLES != [None] and not any(
+                role.name.lower() in BOT_TAGGABLE_ROLES for role in message.author.roles
+            ):
+                return
+
             # Remove the mention from the message
             prompt = message.content.replace(self.bot.user.mention, "")
             # If the message is empty, don't process it
@@ -629,12 +719,7 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
 
             await self.ask_command(
                 message,
-                prompt,
-                False,
-                None,
-                None,
-                None,
-                None,
+                prompt=prompt,
                 from_message_context=True,
             )
 
@@ -689,20 +774,20 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
         )
         embed.add_field(
             name="/gpt ask",
-            value="Ask GPT3 something. Be clear, long, and concise in your prompt. Don't waste tokens.",
+            value="Ask GPT something. Be clear, long, and concise in your prompt. Don't waste tokens.",
             inline=False,
         )
         embed.add_field(
             name="/gpt edit",
-            value="Use GPT3 to edit a piece of text given an instruction",
+            value="Use GPT to edit a piece of text given an instruction",
             inline=False,
         )
         embed.add_field(
-            name="/gpt converse", value="Start a conversation with GPT3", inline=False
+            name="/gpt converse", value="Start a conversation with GPT", inline=False
         )
         embed.add_field(
             name="/gpt end",
-            value="End a conversation with GPT3. You can also type `end` in the conversation.",
+            value="End a conversation with GPT. You can also type `end` in the conversation.",
             inline=False,
         )
         embed.add_field(
@@ -788,18 +873,85 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
         )
         await ctx.respond(embed=embed)
 
+    async def instruction_command(
+        self,
+        ctx: discord.ApplicationContext,
+        mode: str,
+        type: str,
+        instruction: str,
+        instruction_file: discord.Attachment,
+        private: bool,
+    ):
+        """Command to let users set their own system prompt or add one to the channel"""
+
+        await ctx.defer(ephemeral=private)
+
+        if mode == "set" and not (instruction or instruction_file):
+            await ctx.respond(
+                "You must include either an **instruction** or an **instruction file**"
+            )
+            return
+
+        # Check if any of the message author's role names are in CHANNEL_INSTRUCTION_ROLES, if not, continue as user
+        if type == "channel" and mode in ["set", "clear"]:
+            if CHANNEL_INSTRUCTION_ROLES != [None] and not any(
+                role.name.lower() in CHANNEL_INSTRUCTION_ROLES
+                for role in ctx.author.roles
+            ):
+                await ctx.respond(
+                    "You don't have permisson to set the channel instruction. Defaulting to setting a user instruction"
+                )
+                type = "user"
+
+        if instruction_file:
+            bytestring = await instruction_file.read()
+            file_instruction = bytestring.decode("utf-8")
+        if instruction and instruction_file:
+            instruction = f"{file_instruction}\n\n{instruction}"
+        elif instruction_file:
+            instruction = file_instruction
+
+        # If premoderation is enabled, check
+        if PRE_MODERATE:
+            if await Moderation.simple_moderate_and_respond(instruction, ctx):
+                return
+
+        if type == "channel":
+            set_id = ctx.channel.id
+        else:
+            set_id = ctx.user.id
+
+        if mode == "set":
+            self.instructions[set_id] = Instruction(set_id, instruction)
+            await ctx.respond(f"The system instruction is set for **{type}**")
+        elif mode == "get":
+            try:
+                instruction = self.instructions[set_id].prompt
+                embed_pages = await self.paginate_embed(instruction)
+                paginator = pages.Paginator(
+                    pages=embed_pages, timeout=None, author_check=False
+                )
+                await paginator.respond(ctx.interaction)
+            except Exception:
+                await ctx.respond("There is no instruction set")
+        elif mode == "clear":
+            self.instructions.pop(set_id)
+            await ctx.respond(f"The instruction has been removed for **{type}**")
+
     async def ask_command(
         self,
         ctx: discord.ApplicationContext,
         prompt: str,
-        private: bool,
-        temperature: float,
-        top_p: float,
-        frequency_penalty: float,
-        presence_penalty: float,
+        private: bool = False,
+        temperature: float = None,
+        top_p: float = None,
+        frequency_penalty: float = None,
+        presence_penalty: float = None,
         from_ask_action=None,
         from_other_action=None,
         from_message_context=None,
+        prompt_file: discord.Attachment = None,
+        model=None,
     ):
         """Command handler. Requests and returns a generation with no extras to the completion endpoint
 
@@ -815,6 +967,21 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
         is_context = isinstance(ctx, discord.ApplicationContext)
 
         user = ctx.user if is_context else ctx.author
+
+        if not (prompt or prompt_file):
+            await ctx.respond(
+                "You must include either a **prompt** or a **prompt file**"
+            )
+            return
+
+        if prompt_file:
+            bytestring = await prompt_file.read()
+            file_prompt = bytestring.decode("utf-8")
+        if prompt and prompt_file:
+            prompt = f"{file_prompt}\n\n{prompt}"
+        elif prompt_file:
+            prompt = file_prompt
+
         prompt = await self.mention_to_username(ctx, prompt.strip())
 
         if len(prompt) < self.model.prompt_min_length:
@@ -850,6 +1017,7 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
             from_ask_action=from_ask_action,
             from_other_action=from_other_action,
             from_message_context=from_message_context,
+            model=model,
         )
 
     async def edit_command(
@@ -860,7 +1028,6 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
         private: bool,
         temperature: float,
         top_p: float,
-        codex: bool,
     ):
         """Command handler. Requests and returns a generation with no extras to the edit endpoint
 
@@ -870,7 +1037,6 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
             text (str): The text that should be modified
             temperature (float): Sets the temperature override
             top_p (float): Sets the top p override
-            codex (bool): Enables the codex edit model
         """
         user = ctx.user
 
@@ -906,7 +1072,6 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
             overrides=overrides,
             instruction=instruction,
             from_edit_command=True,
-            codex=codex,
             custom_api_key=user_api_key,
         )
 
@@ -934,6 +1099,7 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
         top_p: float,
         frequency_penalty: float,
         presence_penalty: float,
+        use_threads: bool = True,  # Add this parameter
     ):
         """Command handler. Starts a conversation with the bot
 
@@ -969,139 +1135,162 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
             if await Moderation.simple_moderate_and_respond(opener, ctx):
                 return
 
-        # if user.id in self.conversation_thread_owners:
-        #     await ctx.respond(
-        #         "You've already created a thread, end it before creating a new one",
-        #         delete_after=5,
-        #     )
-        #     return
-
-        if private:
-            await ctx.respond(
-                embed=discord.Embed(
-                    title=f"{user.name}'s private conversation with GPT3",
+        if use_threads:
+            if private:
+                embed_title = f"{user.name}'s private conversation with GPT"
+                thread = await ctx.channel.create_thread(
+                    name=embed_title,
+                    auto_archive_duration=60,
+                )
+                target = thread
+            else:
+                embed_title = f"{user.name}'s conversation with GPT"
+                message_embed = discord.Embed(
+                    title=embed_title,
+                    description=f"**Model**: {self.model.model if not model else model}",
                     color=0x808080,
                 )
-            )
-            thread = await ctx.channel.create_thread(
-                name=user.name + "'s private conversation with GPT3",
-                auto_archive_duration=60,
-            )
-        elif not private:
-            message_thread = await ctx.respond(
-                embed=discord.Embed(
-                    title=f"{user.name}'s conversation with GPT3", color=0x808080
+                message_embed.set_thumbnail(url="https://i.imgur.com/asA13vI.png")
+                footer_text = (
+                    "Regular Chat"
+                    if not image_understanding_model.get_is_usable()
+                    else "Regular Chat, Multi-Modal"
                 )
-            )
-            # Get the actual message object for the message_thread
-            message_thread_real = await ctx.fetch_message(message_thread.id)
-            thread = await message_thread_real.create_thread(
-                name=user.name + "'s conversation with GPT3",
-                auto_archive_duration=60,
-            )
+                message_embed.set_footer(
+                    text=footer_text, icon_url="https://i.imgur.com/asA13vI.png"
+                )
+                message_thread = await ctx.send(embed=message_embed)
+                thread = await message_thread.create_thread(
+                    name=user.name + "'s conversation with GPT",
+                    auto_archive_duration=60,
+                )
+                await ctx.respond("Conversation started.")
+                target = thread
+        else:
+            # Check if this current channel is already in a conversation
+            if ctx.channel.id in self.conversation_threads:
+                await ctx.respond(
+                    "There is already a conversation in this channel. Please finish that conversation before starting a new one."
+                )
+                return
 
-        self.conversation_threads[thread.id] = Thread(thread.id)
-        self.conversation_threads[thread.id].model = (
+            # Check if the user is permitted to start a conversation in full channels
+            # check if any of the user role names match CHANNEL_CHAT_ROLES
+            if CHANNEL_CHAT_ROLES and CHANNEL_CHAT_ROLES != [None]:
+                if not any(
+                    role.name.lower() in CHANNEL_CHAT_ROLES for role in ctx.user.roles
+                ):
+                    await ctx.respond(
+                        "You are not permitted to start a conversation in this channel."
+                    )
+                    return
+
+            target = ctx.channel
+            if private:
+                embed_title = f"{user.name}'s private conversation with GPT"
+            else:
+                embed_title = f"{user.name}'s conversation with GPT"
+
+            embed = discord.Embed(title=embed_title, color=0x808080)
+            await ctx.respond(embed=embed)
+
+        self.conversation_threads[target.id] = Thread(target.id)
+        self.conversation_threads[target.id].model = (
             self.model.model if not model else model
         )
 
         # Set the overrides for the conversation
-        self.conversation_threads[thread.id].set_overrides(
+        self.conversation_threads[target.id].set_overrides(
             temperature, top_p, frequency_penalty, presence_penalty
         )
 
-        if opener:
-            opener = await self.mention_to_username(ctx, opener)
-
-        if not opener and not opener_file:
-            user_id_normalized = user.id
-        else:
+        if opener or opener_file:
             user_id_normalized = ctx.author.id
-            if not opener_file:
-                pass
+        else:
+            user_id_normalized = user.id
+        if opener_file:
+            if not opener_file.endswith((".txt", ".json")):
+                opener_file = (
+                    None  # Just start a regular thread if the file fails to load
+                )
             else:
-                if not opener_file.endswith((".txt", ".json")):
+                # Load the file and read it into opener
+                try:
+                    opener_file = re.sub(
+                        ".+(?=[\\//])", "", opener_file
+                    )  # remove paths from the opener file
+                    opener_file = EnvService.find_shared_file(
+                        f"openers{separator}{opener_file}"
+                    )
+                    opener_file = await self.load_file(opener_file, ctx)
+                    try:  # Try opening as json, if it fails it'll just pass the whole txt or json to the opener
+                        opener_file = json.loads(opener_file)
+                        temperature = opener_file.get("temperature", None)
+                        top_p = opener_file.get("top_p", None)
+                        frequency_penalty = opener_file.get("frequency_penalty", None)
+                        presence_penalty = opener_file.get("presence_penalty", None)
+                        self.conversation_threads[target.id].set_overrides(
+                            temperature, top_p, frequency_penalty, presence_penalty
+                        )
+                        if (
+                            not opener
+                        ):  # if we only use opener_file then only pass on opener_file for the opening prompt
+                            opener = opener_file.get("text", "error getting text")
+                        else:
+                            opener = (
+                                opener_file.get("text", "error getting text") + opener
+                            )
+                    except Exception:  # Parse as just regular text
+                        if not opener:
+                            opener = opener_file
+                        else:
+                            opener = opener_file + opener
+                except Exception:
                     opener_file = (
                         None  # Just start a regular thread if the file fails to load
                     )
-                else:
-                    # Load the file and read it into opener
-                    try:
-                        opener_file = re.sub(
-                            ".+(?=[\\//])", "", opener_file
-                        )  # remove paths from the opener file
-                        opener_file = EnvService.find_shared_file(
-                            f"openers{separator}{opener_file}"
-                        )
-                        opener_file = await self.load_file(opener_file, ctx)
-                        try:  # Try opening as json, if it fails it'll just pass the whole txt or json to the opener
-                            opener_file = json.loads(opener_file)
-                            temperature = opener_file.get("temperature", None)
-                            top_p = opener_file.get("top_p", None)
-                            frequency_penalty = opener_file.get(
-                                "frequency_penalty", None
-                            )
-                            presence_penalty = opener_file.get("presence_penalty", None)
-                            self.conversation_threads[thread.id].set_overrides(
-                                temperature, top_p, frequency_penalty, presence_penalty
-                            )
-                            if (
-                                not opener
-                            ):  # if we only use opener_file then only pass on opener_file for the opening prompt
-                                opener = opener_file.get("text", "error getting text")
-                            else:
-                                opener = (
-                                    opener_file.get("text", "error getting text")
-                                    + opener
-                                )
-                        except Exception:  # Parse as just regular text
-                            if not opener:
-                                opener = opener_file
-                            else:
-                                opener = opener_file + opener
-                    except Exception:
-                        opener_file = None  # Just start a regular thread if the file fails to load
 
-        # Append the starter text for gpt3 to the user's history so it gets concatenated with the prompt later
+        # Append the starter text for gpt to the user's history so it gets concatenated with the prompt later
         if minimal or opener_file or opener:
-            self.conversation_threads[thread.id].history.append(
+            self.conversation_threads[target.id].history.append(
                 EmbeddedConversationItem(self.CONVERSATION_STARTER_TEXT_MINIMAL, 0)
             )
         elif not minimal:
-            self.conversation_threads[thread.id].history.append(
+            self.conversation_threads[target.id].history.append(
                 EmbeddedConversationItem(self.CONVERSATION_STARTER_TEXT, 0)
             )
 
         # Set user as thread owner before sending anything that can error and leave the thread unowned
-        self.conversation_thread_owners[user_id_normalized].append(thread.id)
-        overrides = self.conversation_threads[thread.id].get_overrides()
+        self.conversation_thread_owners[user_id_normalized].append(target.id)
+        overrides = self.conversation_threads[target.id].get_overrides()
 
-        await thread.send(f"<@{str(ctx.user.id)}> is the thread owner.")
+        await target.send(f"<@{str(ctx.user.id)}> is the thread owner.")
 
-        await thread.send(
+        await target.send(
             embed=EmbedStatics.generate_conversation_embed(
-                self.conversation_threads, thread, opener, overrides
+                self.conversation_threads, target, opener, overrides
             )
         )
 
         # send opening
         if opener:
-            thread_message = await thread.send(
+            self.conversation_threads[target.id].has_opener = True
+            opener = await self.mention_to_username(ctx, opener)
+            target_message = await target.send(
                 embed=EmbedStatics.generate_opener_embed(opener)
             )
-            if thread.id in self.conversation_threads:
+            if target.id in self.conversation_threads:
                 self.awaiting_responses.append(user_id_normalized)
-                self.awaiting_thread_responses.append(thread.id)
-
                 if not self.pinecone_service:
-                    self.conversation_threads[thread.id].history.append(
+                    self.conversation_threads[target.id].history.append(
                         EmbeddedConversationItem(
                             f"\n{ctx.author.display_name}: {opener} <|endofstatement|>\n",
                             0,
                         )
                     )
+                self.awaiting_thread_responses.append(target.id)
 
-                self.conversation_threads[thread.id].count += 1
+                # ... (no other changes in the middle part of the function)
 
             overrides = Override(
                 overrides["temperature"],
@@ -1112,21 +1301,21 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
 
             await TextService.encapsulated_send(
                 self,
-                thread.id,
+                target.id,
                 opener
-                if thread.id not in self.conversation_threads or self.pinecone_service
+                if target.id not in self.conversation_threads or self.pinecone_service
                 else "".join(
-                    [item.text for item in self.conversation_threads[thread.id].history]
+                    [item.text for item in self.conversation_threads[target.id].history]
                 ),
-                thread_message,
+                target_message,
                 overrides=overrides,
                 user=user,
-                model=self.conversation_threads[thread.id].model,
+                model=self.conversation_threads[target.id].model,
                 custom_api_key=user_api_key,
             )
             self.awaiting_responses.remove(user_id_normalized)
-            if thread.id in self.awaiting_thread_responses:
-                self.awaiting_thread_responses.remove(thread.id)
+            if target.id in self.awaiting_thread_responses:
+                self.awaiting_thread_responses.remove(target.id)
 
     async def end_command(self, ctx: discord.ApplicationContext):
         """Command handler. Gets the user's thread and ends it"""
@@ -1196,12 +1385,7 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
         prompt = await self.mention_to_username(ctx, message.content)
         await self.ask_command(
             ctx,
-            prompt,
-            private=False,
-            temperature=None,
-            top_p=None,
-            frequency_penalty=None,
-            presence_penalty=None,
+            prompt=prompt,
             from_ask_action=prompt,
         )
 
@@ -1225,12 +1409,7 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
 
         await self.ask_command(
             ctx,
-            prompt,
-            private=False,
-            temperature=None,
-            top_p=None,
-            frequency_penalty=None,
-            presence_penalty=None,
+            prompt=prompt,
             from_other_action=from_other_action,
         )
 
@@ -1255,11 +1434,6 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
         await self.ask_command(
             ctx,
             prompt=prompt,
-            private=False,
-            temperature=None,
-            top_p=None,
-            frequency_penalty=None,
-            presence_penalty=None,
             from_other_action=from_other_action,
         )
 
@@ -1285,12 +1459,7 @@ class GPT3ComCon(discord.Cog, name="GPT3ComCon"):
 
         await self.ask_command(
             ctx,
-            prompt,
-            private=False,
-            temperature=None,
-            top_p=None,
-            frequency_penalty=None,
-            presence_penalty=None,
+            prompt=prompt,
             from_other_action=from_other_action,
         )
 
