@@ -1,5 +1,6 @@
 import os
 import sys
+import traceback
 from pathlib import Path
 from typing import Union
 
@@ -43,6 +44,13 @@ class EnvService:
 
         return Path.cwd()
 
+    @staticmethod
+    def save_path():
+        share_dir = os.getenv("SHARE_DIR")
+        if share_dir is not None:
+            return Path(share_dir)
+        return app_root_path()
+
     @staticmethod
     def find_shared_file(file_name):
         share_file_paths = []
@@ -141,7 +149,7 @@ class EnvService:
 
     @staticmethod
     def get_translator_roles():
-        # DALLE_ROLES is a comma separated list of string roles
+        # TRANSLATOR_ROLES is a comma separated list of string roles
         # It can also just be one role
         # Read these allowed roles and return as a list of strings
         try:
@@ -167,7 +175,7 @@ class EnvService:
 
     @staticmethod
     def get_search_roles():
-        # DALLE_ROLES is a comma separated list of string roles
+        # SEARCH_ROLES is a comma separated list of string roles
         # It can also just be one role
         # Read these allowed roles and return as a list of strings
         try:
@@ -219,12 +227,13 @@ class EnvService:
 
     @staticmethod
     def get_index_roles():
-        # GPT_ROLES is a comma separated list of string roles
+        # INDEX_ROLES is a comma separated list of string roles
         # It can also just be one role
         # Read these allowed roles and return as a list of strings
         try:
             index_roles = os.getenv("INDEX_ROLES")
         except Exception:
+            traceback.print_exc()
             index_roles = None
 
         if index_roles is None:
@@ -243,9 +252,65 @@ class EnvService:
         )
         return index_roles
 
+    @staticmethod
+    def get_channel_chat_roles():
+        # CHANNEL_CHAT_ROLES is a comma separated list of string roles
+        # It can also just be one role
+        # Read these allowed roles and return as a list of strings
+        try:
+            cc_roles = os.getenv("CHANNEL_CHAT_ROLES")
+        except Exception:
+            cc_roles = None
+
+        if cc_roles is None:
+            print(
+                "CHANNEL_CHAT_ROLES is not defined properly in the environment file!"
+                "Please copy your server's role and put it into CHANNEL_CHAT_ROLES in the .env file."
+                'For example a line should look like: `CHANNEL_CHAT_ROLES="Gpt"`'
+            )
+            print(
+                "Defaulting to allowing all users to make conversations in full channels..."
+            )
+            return [None]
+
+        cc_roles = (
+            cc_roles.lower().strip().split(",")
+            if "," in cc_roles
+            else [cc_roles.lower()]
+        )
+        return cc_roles
+
+    @staticmethod
+    def get_channel_instruction_roles():
+        # CHANNEL_INSTRUCTION_ROLES is a comma separated list of string roles
+        # It can also just be one role
+        # Read these allowed roles and return as a list of strings
+        try:
+            cc_roles = os.getenv("CHANNEL_INSTRUCTION_ROLES")
+        except Exception:
+            cc_roles = None
+
+        if cc_roles is None:
+            print(
+                "CHANNEL_INSTRUCTION_ROLES is not defined properly in the environment file!"
+                "Please copy your server's role and put it into CHANNEL_INSTRUCTION_ROLES in the .env file."
+                'For example a line should look like: `CHANNEL_INSTRUCTION_ROLES="Gpt"`'
+            )
+            print(
+                "Defaulting to allowing all users to set instructions for channels..."
+            )
+            return [None]
+
+        cc_roles = (
+            cc_roles.lower().strip().split(",")
+            if "," in cc_roles
+            else [cc_roles.lower()]
+        )
+        return cc_roles
+
     @staticmethod
     def get_welcome_message():
-        # WELCOME_MESSAGE is a default string used to welcome new members to the server if GPT3 is not available.
+        # WELCOME_MESSAGE is a default string used to welcome new members to the server if GPT is not available.
         # The string can be blank but this is not advised. If a string cannot be found in the .env file, the below string is used.
         # The string is DMd to the new server member as part of an embed.
         try:
@@ -347,7 +412,7 @@ class EnvService:
                 print(
                     "No user key database path was provided. Defaulting to user_key_db.sqlite"
                 )
-                user_key_db_path = "user_key_db.sqlite"
+                user_key_db_path = EnvService.find_shared_file("user_key_db.sqlite")
             else:
                 # append "user_key_db.sqlite" to USER_KEY_DB_PATH if it doesn't already end with .sqlite
                 if not user_key_db_path.match("*.sqlite"):
@@ -392,6 +457,43 @@ class EnvService:
         except Exception:
             return None
 
+    @staticmethod
+    def get_github_token():
+        try:
+            github_token = os.getenv("GITHUB_TOKEN")
+            return github_token
+        except Exception:
+            return None
+
+    @staticmethod
+    def get_openai_token():
+        try:
+            openai_token = os.getenv("OPENAI_TOKEN")
+            return openai_token
+        except Exception:
+            raise ValueError(
+                "OPENAI_TOKEN is not defined properly in the environment file! The bot cannot start without this token."
+            )
+
+    @staticmethod
+    def get_wolfram_api_key():
+        try:
+            openai_token = os.getenv("WOLFRAM_API_KEY")
+            return openai_token
+        except Exception:
+            print(
+                "WOLFRAM_API_KEY is not defined properly in the environment file! The bot cannot use /internet chat's wolfram functionalities without this"
+            )
+            return None
+
+    @staticmethod
+    def get_openai_organization():
+        try:
+            openai_org = os.getenv("OPENAI_ORGANIZATION")
+            return openai_org
+        except Exception:
+            return None
+
     @staticmethod
     def get_google_search_api_key():
         try:
@@ -400,6 +502,14 @@ class EnvService:
         except Exception:
             return None
 
+    @staticmethod
+    def get_replicate_api_key():
+        try:
+            replicate_key = os.getenv("REPLICATE_API_KEY")
+            return replicate_key
+        except Exception:
+            return None
+
     @staticmethod
     def get_google_search_engine_id():
         try:
@@ -431,3 +541,11 @@ class EnvService:
             return deep_compose_price
         except Exception:
             return 3.00
+
+    @staticmethod
+    def get_google_cloud_project_id():
+        try:
+            google_cloud_project_id = os.getenv("GOOGLE_CLOUD_PROJECT_ID")
+            return google_cloud_project_id
+        except Exception:
+            return None
